// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#ifndef myassert_h_included
#include "myassert.h"
#endif

// Needed by opendir, readdir, and closedir.
#ifndef sys_types_h_included
#include <sys/types.h>
#define sys_types_h_included
#endif

// Needed by opendir, readdir, and closedir.
#ifndef dirent_h_included
#include <dirent.h>
#define dirent_h_included
#endif

// For perror.
#ifndef stdio_h_included
#include <stdio.h>
#define stdio_h_included
#endif

// For strcmp.
#ifndef string_h_included
#include <string.h>
#define string_h_included
#endif

// For std::cerr.
#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

// For dlopen.
#ifndef dlfcn_h_included
#include <dlfcn.h>
#define dlfcn_h_included
#endif

// For set_new_handler.
#ifndef new_h_included
#include <new>
#endif

// For atexit, among others.
#ifndef stdlib_h_included
#include <stdlib.h>
#define stdlib_h_included
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

#ifndef SceneGraph_h_included
#include "SceneGraph.h"
#endif

#ifndef TopConfiguration_h_included
#include "TopConfiguration.h"
#endif

#ifndef TopLevel_h_included
#include "TopLevel.h"
#endif

#ifndef RecursiveSlotValue_h_included
#include "RecursiveSlotValue.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef ActionConfiguration_h_included
#include "ActionConfiguration.h"
#endif

#ifndef Action_h_included
#include "Action.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef Factory_h_included
#include "Factory.h"
#endif

#ifndef AnyFileSceneLoaderConfiguration_h_included
#include "AnyFileSceneLoaderConfiguration.h"
#endif

namespace {
  // From Effective C++, Scott Meyers, page 22.
  void noMoreMemory ()
  {
    std::cerr << "No more memory!" << std::endl;
    abort ();
  }

  // Run the given command.  If it gets an error, use errorChatter to generate
  // an error message, print the message, and exit.
  void runCommand (TopLevel *tl,
		   const FString &commandName, const FString &errorChatter)
  {
    SP<RecursiveSlotValue> rsv =
      NEW (RecursiveSlotValue ("Action", commandName));
    SP<ActionConfiguration> ac =
      dynamic_cast <ActionConfiguration *> (rsv->getConfiguration());
    assert (ac);
    ac->setTopLevel (tl);
    SP<Action> a = dynamic_cast <Action *> (&*rsv->makeIt());
    if (a->isProblem()) {
      std::cerr << errorChatter << ":" << std::endl;
      std::cerr << a->getProblem () << std::endl;
      exit (1);
    }
  }

  void doPerror (const FString &s) {
    perror (&*s);
  }

  void loadPluginsFromDir (const FString &dirName) {
    DIR *dir = opendir (&*dirName);
    if (!dir) {
      doPerror ("Failed to open the directory " + dirName +
                " while looking for plugins");
    } else {
      for (;;) {
        struct dirent *entry = readdir (dir);
        if (!entry) {
          break;
        }
        const char *name = entry->d_name;
        const int namelen = strlen (name);
        // FIXME Change .so to .dll for Microsoft.
        if (namelen > 3 && 0 == strcmp (".so", name+namelen-3)) {
          FString fullName = dirName + "/" + name;
#ifndef NDEBUG
          std::cout << "About to load " << fullName << "." << std::endl;
#endif
          // Should I configure and run LoadPlugin.cpp, or just use dlopen
          // myself?  Configuring and running is too much of a pain, so I'll
          // just use the system routine.
          void *handle = dlopen (&*fullName, RTLD_NOW);
          if (!handle) {
            std::cerr << "Failed to load "<<fullName<<": "<<dlerror ()<<std::endl;
          }
        }
      }
      if (closedir (dir) < 0) {
        doPerror ("Failed to close the directory "+dirName+
                  " while looking for plugins");
      }
    }
  }

  // Load the plugins specified on the search path.
  // FIXME Microsoft OS's want ";" to be the separator on the path, not ":".
  void loadPlugins (FString path) {
    int start = 0;
    while (start < path.size ()) {
      int end = start;
      while (end < path.size() && path [end] != ':') {
        end++;
      }
      loadPluginsFromDir (path.substring (start, end - start));
      start = end + 1;
    }
  }
}

int main(int argc, const char **argv) {
  MemoryUtil::allLeaksAreInteresting ();
  std::cout << "Fungimol version " VERSION << std::endl
       << "Copyright 2000 Tim Freeman" << std::endl
       << "Fungimol is free software, covered by the Gnu Library General "
       << "Public License" << std::endl
       << "and you are welcome to change it and/or distribute copies of it "
       << "under certain" << std::endl
       << "conditions." << std::endl
       << "There is absolutely no warranty for Fungimol." << std::endl
       << "Read the file " DOCHOME "/COPYING.txt for the copying conditions "
       << std::endl
       << "and details about the lack of warranty." << std::endl
       << "Visit the URL file:" DOCHOME "/index.html for other documentation."
       << std::endl;
  SP<TopConfiguration> theTop;
  {
    // Next line has to change for a Windows port.
    FString WindowSystemName = "X";
    theTop = NEW (TopConfiguration (WindowSystemName + "TopLevel"));
  }
  std::set_new_handler (noMoreMemory);
  bool firstTime = true;
  // This loop goes around once each time an Action tells theTopLevel->mainLoop
  // to exit.  This typically happens when the Action changes the configuration
  // for making the TopLevel enough that a new TopLevel needs to be made, which
  // is typically when a new window system is specified.
  SP<TopLevel> theTopLevel = 0;
  for (;;) {
    {
      // Preserve the scene graph and the top level configuration from
      // one incarnation of the toplevel to the next.
      SP<SceneGraph> theSceneGraph = 0;
      if (theTopLevel) {
	theSceneGraph = theTopLevel->getSceneGraph ();
	theTop = theTopLevel->getConfiguration ();
      }
      // Make sure the lifetime of the old top level doesn't overlap the
      // lifetime of the new one.  The problem here is that creating and
      // destroying toplevel's has to reset the object drawer connections; it's
      // important that the previous toplevel resets the object drawer
      // connections befroe the new toplevel sets them.
      theTopLevel = 0;
      theTopLevel =
	dynamic_cast <TopLevel *> (&*(theTop->getTopLevelSlot()->makeIt()));
      assert (theTopLevel);
      theTopLevel->setSceneGraph (theSceneGraph);
      theTopLevel->setConfiguration (theTop);
    }
    if (firstTime) {
      firstTime = false;
      // Load the plugins after creating the top level, so they can bind keys
      // if they want to.  Load them before loading the initial scene, so they
      // can define scene loaders that can then be used to load the initial
      // scene.
      {
        const char *pluginpath = getenv ("FUNGIMOL_PATH");
        if (pluginpath) loadPlugins (pluginpath);
      }
      loadPlugins (DEFAULT_FUNGIMOL_PATH);
      // Make a new scene.
      if (argc == 1) {
	// Do nothing
      } else if (argc == 2) {
	// The scene generator for the new scene should read from the file with
	// the name given on the command line.
	SP<RecursiveSlotValue> rsv =
	  NEW (RecursiveSlotValue ("SceneLoader", "AnyFileSceneLoader"));
	SP<AnyFileSceneLoaderConfiguration> aslc =
	  dynamic_cast <AnyFileSceneLoaderConfiguration *>
	  (&*rsv->getConfiguration ());
	assert (argv [1]);
	// Not much point in having the newly-loaded stuff selected.
	aslc->setSelect (false);
	aslc->setFileName (argv[1]);
	aslc->setTakeDefaults (true);
	theTopLevel->getConfiguration()->setSceneLoaderSlot (rsv);
      } else {
	std::cerr << "Usage: " << argv[0] << " [<filename>]" << std::endl;
	exit (1);
      }
      runCommand (theTopLevel, "NewScene",
		  "Could not create the initial scene");
    }
    // Do the ViewScene if this is the first time or if we've just made a new
    // TopLevel.
    runCommand (theTopLevel, "ViewScene",
                "Failed to view the initial scene");
    if (!theTopLevel->mainLoop ()) break;
  }
#ifndef NDEBUG
  theTopLevel = 0;
  theTop = 0;
  MemoryUtil::listLeaks();
#endif
}
