// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#ifndef math_h_included
#include <math.h>
#define math_h_included
#endif

#ifndef TypedFactory_h_included
#include "TypedFactory.h"
#endif

#ifndef AtomConfiguration_h_included
#include "AtomConfiguration.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef MemoryUtil_h_included
#include "MemoryUtil.h"
#endif

#ifndef AtomInfo_h_included
#include "AtomInfo.h"
#endif

#ifndef FFloat_h_included
#include "FFloat.h"
#endif

#ifndef PhysicsObjectInfo_h_included
#include "PhysicsObjectInfo.h"
#endif

#ifndef FloatUtil_h_included
#include "FloatUtil.h"
#endif

#ifndef Vec3_h_included
#include "Vec3.h"
#endif

#ifndef AtomCache_h_included
#include "AtomCache.h"
#endif

#ifndef BoringAtom_h_included
#include "BoringAtom.h"
#endif

namespace {
  // If a design atom is not linked to another design atom, it will try to push
  // that atom away to designAtomRadius angstroms.  If this is too big, then
  // neighbor computation is slow.  If this is too small, then you don't get
  // the functionality you want because they'll clump too much.
  // 
  // DesignAtom's don't have momentum.  The force is interpreted directly as
  // the derivative of position, so things don't bounce around too much.
  // FIXME If I set designAtomRadius to 6, it doesn't work at all.
  // Things are too far apart, probably due to the maximum attractive force 
  const FFloat designAtomRadius = 3;
  // Equilibrium bond lengths for design atoms.  We want these to be close
  // enough so that if we do a minimization pass with Brenner's code, it will
  // adjust things so they are right.  In general, too small numbers are better
  // than too large numbers, since small distances will cause atoms to push
  // apart to the correct distance, but large distances may cause things to
  // fail to interact at all and then we get the wrong minimum.
  const FFloat designHH = 0.75;
  const FFloat designHC = 1.12;
  const FFloat designHN = 1.01;
  const FFloat designHO = 0.94;
  const FFloat designCC = 1.42;
  const FFloat designCN = 1.45;
  const FFloat designCO = 1.41;
  const FFloat designCS = 1.80;
  const FFloat designOP = 1.61;
  const FFloat designNP = 1.66;
  // headRoom is the distance that we can permit our force field to extend
  // beyond the equilibrium distance.  We want a safe distance for all
  // equilibrium distances, which means we want to subtract out the largest
  // equilibrium distance, which happens to be designCC.
  // FIXME The largest equilibrium distance is now *designCN*.
  // However I don't have time right now to carefully exercise this after
  // tweaking the numbers, so I'll leave headRoom the same for now.  (Tim
  // Freeman  8 Jul 2000.)
  const FFloat headRoom = designAtomRadius - designCC;
  const int designDistTabSize = 17;
  const FFloat designDistTab [designDistTabSize] [designDistTabSize] =
  {{0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, designHH, 0, 0, 0, 0, designHC, designHN, designHO, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, designHC, 0, 0, 0, 0, designCC, designCN, designCO, 0, 0, 0, 0, 0, 0, 0, designCS},
   {0, designHN, 0, 0, 0, 0, designCN, 0, 0, 0, 0, 0, 0, 0, 0, designNP, 0},
   {0, designHO, 0, 0, 0, 0, designCO, 0, 0, 0, 0, 0, 0, 0, 0, designOP, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, 0, designNP, designOP, 0, 0, 0, 0, 0, 0, 0, 0},
   {0, 0,        0, 0, 0, 0, designCS, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};
  bool didInit = false;
  // Next table will have the square of the distance between 
  FFloat maxDist  [designDistTabSize] [designDistTabSize];
  // Multiply all forces by this much.  Compensating for the fact that I
  // changed the time step to be 50 times larger and am using a numerical
  // analysis technique that is somewhat less accurate.  To test, make a
  // string of 8 carbons linked together; if they have to fight with each other
  // to find an equilibrium, then you have to increase the denominator.  1/200
  // is too small if the timestep is 0.5 and the integration technique is
  // AB3.  Tim Freeman  7 Nov 2000.
  const FFloat timeRatio = 1/300.0;
  class DesignAtom
    : public BoringAtom
  {
  public:
    DesignAtom (AtomConfiguration *b, const Factory *f)
      : BoringAtom (b, f)
    {
      if (!didInit) {
	didInit = true;
	for (int i = 0; i < designDistTabSize; i++) {
	  for (int j = 0; j < designDistTabSize; j++) {
	    if (0 == designDistTab [i] [j]) {
	      maxDist [i] [j] = 0;
	    } else {
	      const FFloat f = designDistTab [i] [j] + headRoom;
	      maxDist [i] [j] = f * f;
	    }
	  }
	}
      }
    }
    FFloat physicsBoundingSphereRadius () const {
      return designAtomRadius;
    };
    static const char *const staticClassId;
    // We've adjusted dist so that the equilibrium distance is 0.  Return a
    // positive force that has a continuous fourth derivative (to make
    // RungeKutta happy) and peters down to 0 at dist = headRoom.
    // According to gcov, about 90% of the time this returns 0.
    static FFloat compForce (const FFloat dist) {
      const FFloat headDist = headRoom - dist;
      assert (headDist >= 0);
      const FFloat d2 = headDist * headDist;
      const FFloat d4 = d2 * d2;
      return d4 * headDist * timeRatio;
    }
    void doPhysics (const FFloat *state,
		    FFloat *deriv,
		    int myIndex,
		    const Dynavec<PhysicsObjectInfo> &neighbors,
		    const Dynavec<PhysicsObjectInfo> &links) const;
    bool isRelevantNeighbor (const PhysicsObject *o, int myIndex, int neighborIndex) const;
    const char *classId () const {
      return staticClassId;
    }
    bool isInstanceOf (const char *classId) const {
      return classId == staticClassId || BoringAtom::isInstanceOf (classId);
    }
  };
  // If (it's a DesignAtom and it's index is greater than our own,
  // or it is not a DesignAtom, and it's some other kind of atom),
  // and it's atomic symbol is something we know how to deal with, then we want
  // it on the neighbor list.
  bool DesignAtom::isRelevantNeighbor (const PhysicsObject *o, int myIndex, int neighborIndex)
    const
  {
    const bool isDA = o->classId() == DesignAtom::staticClassId;
    if ((isDA && (neighborIndex > myIndex)) ||
	((!isDA) && o->isInstanceOf (BoringAtom::staticClassId))) {
      const BoringAtom *da = static_cast <const BoringAtom *> (o);
      const int toSym = da->getSymbolNumber ();
      assert (toSym > 0);
      if (toSym < designDistTabSize) {
	return true;
      } else {
	return false;
      }
    } else {
      return false;
    }
  }
  void DesignAtom::doPhysics (const FFloat *state,
			      FFloat *deriv,
			      int myIndex,
			      const Dynavec<PhysicsObjectInfo> &neighbors,
			      const Dynavec<PhysicsObjectInfo> &links) const {
    const int fromSym = getSymbolNumber ();
    const Vec3 myCenter = center (state + myIndex);
    assert (fromSym > 0);
    if (fromSym >= designDistTabSize) return;
    // Essentially all the time happens in the following loop, because we have
    // hundreds of nearby atoms but few links.
    for (int i = 0; i < neighbors.size(); i++) {
      PhysicsObjectInfo poi = neighbors [i];
      assert (poi.m_object);
#ifndef NDEBUG
      const bool isDA = poi.m_object->classId() == DesignAtom::staticClassId;
#endif
      assert ((isDA & (poi.m_stateIndex > myIndex)) ||
	      ((!isDA) &&
	       poi.m_object->isInstanceOf (BoringAtom::staticClassId)));
      const BoringAtom *da =
	static_cast <const BoringAtom *> (poi.m_object);
      const int toSym = da->getSymbolNumber ();
      assert (toSym > 0);
      // Vector from me to the other guy.
      const Vec3 vec = BoringAtom::center (state + poi.m_stateIndex) -
	myCenter;
      const FFloat distSquared = vec.lengthSquared();
      const FFloat maxD = maxDist [fromSym] [toSym];
      if ((distSquared < maxD) && (distSquared >= 0.01 * 0.01)) {
	// The purpose of maxD is to avoid taking the following square root
	// most of the time.
	const FFloat dist = sqrt (distSquared);
	const FFloat equilibriumDist = designDistTab [fromSym] [toSym];
	const FFloat force = compForce (dist - equilibriumDist);
	assert (force > 0);
	const Vec3 forceVec = vec * (force / dist);
	setCenter (deriv+myIndex, center (deriv+myIndex)-forceVec);
	if (poi.m_object->classId() == DesignAtom::staticClassId) {
	  setCenter (deriv+poi.m_stateIndex,
		     center (deriv+poi.m_stateIndex)+forceVec);
	}
      }
    }
    for (int i = 0; i < links.size(); i++) {
      PhysicsObjectInfo poi = links [i];
      // Note that we are looping over *links* here, not neighbors.  Thus the
      // isRelevantNeighbor check doesn't apply, so we have to do the same
      // computation again here.  This isn't too bad because an atom typically
      // has many nonbonded neighbors (so isRelevantNeighbor wins big)
      // and only very few bonded neighbors (so this code loses small).
      // 
      // poi.m_object can be null.  Here's the scenario: create an atom, make
      // another atom, link them, delete one of them.  The link index that the
      // other one has in its list has to become -1 because we support links
      // with meaning based on their position in the link list, and the
      // corresponding m_object ought to be 0 because there's no other good
      // value.
      if (0 != poi.m_object) {
        const bool isDA = poi.m_object->classId() == DesignAtom::staticClassId;
	if ((isDA & (poi.m_stateIndex > myIndex)) ||
	    ((!isDA) & poi.m_object->isInstanceOf
             (BoringAtom::staticClassId))) {
	  const BoringAtom *da =
            static_cast <const BoringAtom *> (poi.m_object);
	  const int toSym = da->getSymbolNumber ();
	  assert (toSym > 0);
	  if (toSym < designDistTabSize) {
	    const FFloat equilibriumDist = designDistTab [fromSym] [toSym];
	    if (equilibriumDist > 0) {
	      // Vector from me to the other guy.
	      const Vec3 vec =
		DesignAtom::center (state + poi.m_stateIndex) - myCenter;
	      const FFloat dist = vec.length();
              // If the distance is very small, then don't do anything, so we
              // don't get a zero divide when two atoms are placed squarely on
              // top of each other.
	      if (dist > 0.01) {
		// Use min here because the distance can be quite large, and we
		// don't want to get astronomical velocities and broken
		// numerical integration in that case.
		const FFloat force =
		  FloatUtil::min ((FFloat)(30.0 * timeRatio),
                                  compForce (equilibriumDist - dist));
		assert (force >= 0);
		const Vec3 forceVec = vec * (force / dist);
		setCenter (deriv+myIndex, center (deriv+myIndex)+forceVec);
		if (isDA) {
                  // If the other guy is a DesignAtom too, then change his
                  // velocity too.  The cost is an if statement and the benefit
                  // is that we don't have to consider the other guy separately
                  // and recompute the force then.
		  setCenter (deriv+poi.m_stateIndex,
			     center (deriv+poi.m_stateIndex)-forceVec);
		}
	      }
	    }
	  }
	}
      }
    }
  }
  const char *const DesignAtom::staticClassId = "DesignAtom";
  class DesignAtomFactory
    : TypedFactory <AtomConfiguration, DesignAtom>
  {
    static SP<AtomCache <DesignAtom> > s_cache;
    static void deallocate () {
      if (s_cache) {
	s_cache = 0;
      }
    }
    static const bool s_useless;
  public:
    DesignAtomFactory ()
      : TypedFactory <AtomConfiguration, DesignAtom> ("DesignAtom")
    {}
    SP<AtomConfiguration> typedDefaultConfiguration () const {
      return NEW (AtomConfiguration ());
    }
    SP<DesignAtom> makeIt (AtomConfiguration *ai) const {
      SP<DesignAtom> &da = s_cache->cacheref (ai);
      if (!da) {
	da = NEW (DesignAtom (ai, this));
      }
      return da;
    }
  };
  SP<AtomCache<DesignAtom> > DesignAtomFactory::s_cache =
  NEW (AtomCache <DesignAtom> ());
  const bool DesignAtomFactory::s_useless =
  (FactoryTable::store("Atom", NEW (DesignAtomFactory ())),
   MemoryUtil::registerDeallocator (deallocate),
   true);
}
