// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#include "HashTable.h"

// Nothing here but a test case.  It uses FString, which is in another
// package.
#ifdef UNIT_TEST
// Compilation command:
// g++ -Wall -Wwrite-strings -Werror -pipe -felide-constructors -g  -I/home/tim/fungimol/Util -I/home/tim/fungimol/FString -I. -DUNIT_TEST /home/tim/fungimol/Hashtable/HashTable.cpp /home/tim/fungimol/FString/*.o /home/tim/fungimol/Hashtable/*.o /home/tim/fungimol/Util/*.o -o test

#ifdef NDEBUG
#warn This test case makes no sense if compiled with -DNDEBUG.
#endif

#ifndef MemoryUtil_h_included
#include "MemoryUtil.h"
#endif

#ifndef stdio_h_included
#include <stdio.h>
#define stdio_h_included
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#define MAXLEN 200

#ifndef Refcount_h_included
#include "Refcount.h"
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

// Need next one for growths, probes, and collisions.
#ifndef HashTableImpl_h_included
#include "HashTableImpl.h"
#endif

class RefcountedInteger : public Refcount {
  int m_int;
public:
  RefcountedInteger (int i) : m_int (i) {}
  int getInt () const { return m_int; }
  void setInt (int i) { m_int = i; }
  operator int () const { return m_int; }
};

int main (int argc, char **argv)
{
  char *filename;
  FILE *fp;
  if (2 != argc) {
    printf ("Should pass exactly one file argument, for now.\n");
    exit (1);
  }
  filename = argv [1];
  fp = fopen (filename, "r");
  if (0 == fp) {
    printf ("Failed to open %s.\n", filename);
    exit (2);
  }
  {
    HashTable <FString, RefcountedInteger> tab;
    int lineTotal = 0;
    int repeatedLineTotal = 0;
    
    /* First time, make a table mapping the text of each line in */
    /* the file to the line number.  If the line is repeated, the */
    /* first line number goes in the table.  lineTotal is the sum */
    /* of the line numbers in the table and repeatedLineTotal is */
    /* the sum of the line numbers of the repetitions of the */
    /* lines. */
    char buf [MAXLEN];
    char *cp;
    {
      int lineno = 0;
      for (;;) {
	cp = fgets (buf, MAXLEN, fp);
	if (0 == cp) break;
	lineno++;
	if (tab.load (FString(buf))) {
	  repeatedLineTotal += lineno;
	} else {
	  lineTotal += lineno;
	  tab.store (buf, NEW (RefcountedInteger (lineno)));
	}
      }
      printf ("Finished storing %d lines.\n", lineno);
    }
    rewind (fp);
    {
      int lineno = 0;
      int lineTotal2 = 0;
      for (;;) {
	SP<RefcountedInteger> oldLine;
	cp = fgets (buf, MAXLEN, fp);
	if (0 == cp) break;
	lineno++;
	oldLine = tab.load (buf);
	assert (oldLine);
	assert ((*oldLine) <= lineno);
	if ((*oldLine) == lineno) {
	  lineTotal2 += lineno;
	} else {
	  repeatedLineTotal -= lineno;
	}
      }
      assert (lineTotal2 == lineTotal);
      assert (0 == repeatedLineTotal);
    }
    {
      int stat = fclose (fp);
      if (0 != stat) {
	printf ("Failed to close %s.\n", filename);
	exit (3);
      }
    }
    HashTable <FString, RefcountedInteger>::HashTableIterator i =
      tab.iterator ();
    for (;;) {
      if (i.isDone()) break;
      lineTotal -= *(tab.load (i.currentItem()));
      i.next ();
    }
    assert (0 == lineTotal);
  }
  printf ("%d growths; %d probes; %d collisions\n",
	  HashTableImpl::growths,
	  HashTableImpl::probes,
	  HashTableImpl::collisions);
  listLeaks ();
  return 0;
}
#endif
