// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#include "DispatcherConfiguration.h"

#ifndef Configuration_h_included
#include "Configuration.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef StringSlotValue_h_included
#include "StringSlotValue.h"
#endif

#ifndef RecursiveSlotValue_h_included
#include "RecursiveSlotValue.h"
#endif

#ifndef MemoryUtil_h_included
#include "MemoryUtil.h"
#endif

#ifndef VectorSlotValue_h_included
#include "VectorSlotValue.h"
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef Configurable_h_included
#include "Configurable.h"
#endif

#ifndef Event_h_included
#include "Event.h"
#endif

#ifndef ButtonSlotValue_h_included
#include "ButtonSlotValue.h"
#endif

#ifndef VecUtil_h_included
#include "VecUtil.h"
#endif

#ifndef ConfigurationSlotValue_h_included
#include "ConfigurationSlotValue.h"
#endif

namespace {
  typedef Event::Button Button;
  class KeyBinding
    : public Configuration
  {
    static FString keyName () {
      return "Key";
    }

    static FString factoryName () {
      return "What it should do";
    }
    FString m_protocol;
  public:
    // Unfortunately, the factory protocol requires us to be able to make a
    // default configuration without knowing anything, so we can't pass the
    // event into the KeyBinding constructor.  Thus we have to separate the
    // constructor from the initializeSlots method.
    KeyBinding (const FString &protocol)
      : m_protocol (protocol)
    {}
    void initializeSlots (SP<const Event> event) {
      initializeSlot (keyName(), NEW (ButtonSlotValue (event)));
      initializeSlot (factoryName(),
		      NEW (RecursiveSlotValue (m_protocol, "DoNothing")));
    }    
    KeyBinding (const KeyBinding &kb)
      : Configuration (kb)
    {}
    Button getKey () const {
      return ButtonSlotValue::getSlot (this, keyName ());
    }
    void setKey (Button ch) {
      ButtonSlotValue::setSlot (this, keyName (), ch);
    }
    SP<RecursiveSlotValue> getBinding () {
      return RecursiveSlotValue::getSlotValue (this, factoryName());
    }
    SP<const RecursiveSlotValue> getBinding () const {
      return RecursiveSlotValue::getSlotValue (this, factoryName());
    }
    void setBinding (SP<RecursiveSlotValue> rsv) {
      RecursiveSlotValue::setSlotValue (this, factoryName(), rsv);
    }
    bool hasTheSameTypeAs (const Configuration *c) const {
      return 0 != dynamic_cast <const KeyBinding *> (c);
    }
    SP<Configuration> copy () const {
      return NEW (KeyBinding (*this));
    }
    FString unParse () const {
      SP<const SlotValue> slot = getSlot (keyName ());
      assert (slot);
      FString keyName = slot->unParse ();
      // I think the next three lines could have been replaced by the following
      // one, but g++ doesn't like it, and I couldn't quickly figure out why.
      // FString keyName = getSlot (keyName ())->unParse ();
      return keyName + " is bound to " + getBinding()->unParse();
    }
  };

  // Return the binding for ch if it is found, otherwise null.
  SP<KeyBinding> getKeyBinding (SP<VectorSlotValue> bindings, Button key)
  {
    for (int i = 0; i < bindings->size(); i++) {
      SP<SlotValue> sv = (*bindings)[i];
      SP<ConfigurationSlotValue> crsv =
	dynamic_cast <ConfigurationSlotValue *> (&*sv);
      SP<KeyBinding> kb =
	dynamic_cast <KeyBinding *> (&*crsv->getConfiguration());
      assert (kb);
      if (key == kb->getKey()) {
	return kb;
      }
    }
    return 0;
  }

  SP<ConfigurationSlotValue> newKeyBinding (SP<const Event> event,
                                            const FString &protocol) {
    SP<KeyBinding> kbconf = NEW (KeyBinding (protocol));
    kbconf->initializeSlots (event);
    SP<ConfigurationSlotValue> result =
      NEW (ConfigurationSlotValue (*kbconf, "key binding"));
    return result;
  }
} // namespace

struct DispatcherConfiguration::DispatcherConfigurationData {
  // An event of the approriate class, which we can use to parse and unparse
  // buttons and so forth.
  SP<const Event> m_event;
  // The specific superclass of Action that we are willing to let the user put
  // into this dispatch table.  Typically "KeyboardMouseAction" or
  // "SpaceOrbAction".
  FString m_protocol;
  // All of the shift keys.
  // FIXME Could allow the user to configure the set of shift keys.
  Dynavec <Event::Button> m_shiftButtons;
  FString bindingSlotName (Button b) {
    if (m_event->maxButton() == b) {
      return "Bindings with no shift keys";
    } else {
      return
	("Bindings when " + m_event->unParseButton (b)
	 + " is down");
    }
  }
};

DispatcherConfiguration::DispatcherConfiguration
(SP<const Event> event,
 const FString &protocol,
 const Dynavec <Button> &shiftButtons)
  : m_data (NEW (DispatcherConfigurationData ()))
{
  m_data->m_event = event;
  m_data->m_protocol = protocol;
  // Add maxButton () to the beginning of the list of shift keys, so the
  // un-shifted case isn't a special case.  Add it to the beginning so the
  // simple keys come out first when we list all keys.
  m_data->m_shiftButtons.push (event->maxButton ());
  for (int i = 0; i < shiftButtons.size(); i++) {
    m_data->m_shiftButtons.push (shiftButtons [i]);
  }
  for (int i = 0; i < m_data->m_shiftButtons.size(); i++) {
    initializeSlot (m_data->bindingSlotName (m_data->m_shiftButtons [i]),
		    NEW (VectorSlotValue (&*newKeyBinding (&*event,
							   protocol))));
  }
}

DispatcherConfiguration::~DispatcherConfiguration () {
  assert (m_data);
  delete m_data;
  m_data = 0;
}

DispatcherConfiguration::DispatcherConfiguration
(const DispatcherConfiguration &dc)
  : ActionConfiguration (dc),
    m_data (NEW (DispatcherConfigurationData (*dc.m_data)))
{}

SP<Configuration> DispatcherConfiguration::copy () const {
  return NEW (DispatcherConfiguration (*this));
}

bool DispatcherConfiguration::hasTheSameTypeAs (const Configuration *c)
  const
{
  return 0 != dynamic_cast <const DispatcherConfiguration *> (c);
}

void DispatcherConfiguration::bindButton (Button key, RecursiveSlotValue *a_f)
{
  bindButton (key, a_f, m_data->m_event->maxButton ());
}

void DispatcherConfiguration::bindButton (Button key, RecursiveSlotValue *a_f,
					  Button shift) {
  SP<RecursiveSlotValue> f = a_f;
  FString name;
  if (!VecUtil::find (shift, m_data->m_shiftButtons)) {
    assert (0 && "Bad shift value in bindKey");
    // Ignore the bad binding.
  } else {
    name = m_data->bindingSlotName (shift);
    SP<VectorSlotValue> vsv =
      dynamic_cast <VectorSlotValue *> (&*getSlot (name));
    SP<KeyBinding> binding = getKeyBinding (vsv, key);
    if (binding) {
      binding->setBinding (f);
    } else {
      int where = vsv->size();
      vsv->extendTo (where + 1);
      SP<ConfigurationSlotValue> newBinding =
	newKeyBinding (m_data->m_event, m_data->m_protocol);
      SP<KeyBinding> kb =
	dynamic_cast <KeyBinding *> (&*newBinding->getConfiguration());
      assert (kb);
      kb->setKey (key);
      kb->setBinding (f);
      vsv->setValue (where, &*newBinding);
    }
  }
}

void DispatcherConfiguration::bindButton
(Button key, const FString &factoryName, Button shift) {
  bindButton (key,
	      NEW (RecursiveSlotValue (m_data->m_protocol, factoryName)),
	      shift);
}

void DispatcherConfiguration::bindButton
(Button key, const FString &factoryName) {
  bindButton (key, factoryName, m_data->m_event->maxButton ());
}

SP<RecursiveSlotValue> DispatcherConfiguration::getBinding
(SP<const Event> event)
{
  // Should only call us when the mouse moved or a key went up.
  // Calling this when a key goes up is weird.  
  assert (event->getButton() == m_data->m_event->pointerMotionButton () ||
	  event->isDown());
  const Button maxButton = m_data->m_event->maxButton ();
  assert (event->maxButton () == maxButton);
  bool found = false;
  // If we don't initialize whichShift here, the compiler thinks it may be read
  // before written.  This isn't true, but no use arguing.
  Button whichShift = 0;
  for (int i = 0; i < m_data->m_shiftButtons.size(); i++) {
    // If the shift button at position i is down (or it is maxButton, which
    // means we're taking the non-shifted interpretation), and all of the other
    // shift buttons are up, then we've discovered which binding map to use.
    const Button thisButton = m_data->m_shiftButtons [i];
    if ((thisButton == maxButton) || event->isButtonDown (thisButton)) {
      // Okay, this button is down, so if all of the other shift buttons are
      // up, we're done.  But be sure not to query whether maxButton is up, or
      // whether this button is up.
      bool otherButton = false;
      for (int j = 0; j < m_data->m_shiftButtons.size(); j++) {
	const Button thatButton = m_data->m_shiftButtons [j];
	if (thatButton != thisButton &&
	    thatButton != maxButton &&
	    event->isButtonDown (thatButton)) {
	  otherButton = true;
	}
      }
      if (!otherButton) {
	found = true;
	whichShift = thisButton;
      }
    }
  }
  if (! found) {
    return 0;
  } else {
    FString name = m_data->bindingSlotName (whichShift);
    SP<VectorSlotValue> rsv =
      dynamic_cast <VectorSlotValue *> (&*getSlot (name));
    SP<KeyBinding> binding = getKeyBinding (rsv, event->getButton ());
    if (binding) {
      return binding->getBinding ();
    } else {
      return 0;
    }
  }
}

const Dynavec <Button> &DispatcherConfiguration::getShiftButtons () const {
  return m_data->m_shiftButtons;
}


std::ostream &operator<< (std::ostream &o,
		     const DispatcherConfiguration &dc)
{
  Dynavec <Dynavec <FString> > tableau;
  tableau.push (Dynavec <FString> ());
  tableau[0].push ("Key");
  tableau[0].push ("Action");
  for (int j = 0; j < dc.m_data->m_shiftButtons.size(); j++) {
    const Button shiftButton = dc.m_data->m_shiftButtons [j];
    FString slotName = dc.m_data->bindingSlotName (shiftButton);
    FString prefix;
    if (dc.m_data->m_event->maxButton () == shiftButton) {
      // If there's no shift key, we have an empty prefix.
      prefix = "";
    } else {
      // Otherwise, if (for example) the meta key is pushed, the prefix is
      // "Meta ".
      prefix = dc.m_data->m_event->unParseButton (shiftButton) + " ";
    }
    SP<const VectorSlotValue> rsv =
      dynamic_cast <const VectorSlotValue *> (&*dc.getSlot (slotName));
    for (int i = 0; i < rsv->size(); i++) {
      SP<const ConfigurationSlotValue> crsv =
	dynamic_cast <const ConfigurationSlotValue *> (&*((*rsv)[i]));
      assert (crsv);
      SP<const KeyBinding> kb =
	dynamic_cast <const KeyBinding *>  (&*crsv->getConfiguration());
      assert (kb);
      Button key = kb->getKey();
      tableau.push (Dynavec <FString> ());
      // If the meta-a key is pushed, then prefix will be "Meta ", so we
      // will put "Meta a" on the left column of the table.
      tableau[tableau.size()-1].push (prefix +
				      dc.m_data->m_event->unParseButton (key));
      tableau[tableau.size()-1].push (kb->getBinding()->unParse());
    }
  }
  StreamUtil::printTableau (o, tableau);
  return o;
}

