// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#ifndef TheObjectDrawerFactory_h_included
#include "TheObjectDrawerFactory.h"
#endif

#ifndef MemoryUtil_h_included
#include "MemoryUtil.h"
#endif

#ifndef ObjectDrawerConnection_h_included
#include "ObjectDrawerConnection.h"
#endif

#ifndef ConstantRecursiveSlotValue_h_included
#include "ConstantRecursiveSlotValue.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef ObjectDrawerConfiguration_h_included
#include "ObjectDrawerConfiguration.h"
#endif

#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef Refcount_h_included
#include "Refcount.h"
#endif

class TheObjectDrawerFactory::TheObjectDrawerFactoryData : public Refcount {
public:
  Dynavec <ObjectDrawerConnection *> m_connections;
  SP<ConstantRecursiveSlotValue> m_odf;
  Dynavec <FString> m_objectDrawers;
  static void makeIt ();
  static void deallocate ();
};

// There was a bug here.  We had deallocators that were indirectly calling
// deRegisterConnection after TheObjectDrawerFactory had already been
// deallocated.  The solution I chose here is to make TheObjectDrawerFactory
// reference counted, and to make everyone who wants to be able to
// deRegisterConnection themselves retain a pointer to it.  This way, we can
// zero out the static pointer to the factory without breaking
// deRegisterConnection.  If they try to make a new PhysicsObject that that
// time they'll lose, but there's no good reason to make new objects while
// deallocating anyway.  Tim Freeman  1 Nov 2000.
namespace {
  // We cannot ensure that our static initializers get run before the static
  // initializers of classes that use TheObjectDrawerFactory, so we can't use a
  // static initializer here.  I think that if we make it a SP then it will
  // implicitly get a static initializer, so it has to be a plain pointer.
  TheObjectDrawerFactory *s_me;
};

void TheObjectDrawerFactory::TheObjectDrawerFactoryData::deallocate () {
  if (s_me) {
    s_me->deref ();
    if (0 == s_me->count()) {
      delete s_me;
    }
  }
  s_me = 0;
}

void TheObjectDrawerFactory::TheObjectDrawerFactoryData::makeIt () {
  if (!s_me) {
      s_me = NEW (TheObjectDrawerFactory ());
      s_me->ref();
      MemoryUtil::registerDeallocator (deallocate);
  }
}

TheObjectDrawerFactory::~TheObjectDrawerFactory () {
  assert (m_data);
  delete m_data;
  m_data = 0;
}

TheObjectDrawerFactory::TheObjectDrawerFactory ()
  : m_data (NEW (TheObjectDrawerFactoryData ()))
{
}

void TheObjectDrawerFactory::giveObjectDrawer
(SP<ObjectDrawerConnection> odconn)
{
  TheObjectDrawerFactoryData::makeIt ();
  assert (odconn);
  SP<ConstantRecursiveSlotValue> odf = s_me->m_data->m_odf;
  if (odf) {
    SP<ObjectDrawerConfiguration> odconf =
      dynamic_cast <ObjectDrawerConfiguration *> (&*(odf->getConfiguration()));
    odconf->setClassName (odconn->getObjectDrawingClass());
    odconf->setPhysicsObjectConfiguration (odconn->getConfiguration());
    SP<ObjectDrawer> od = dynamic_cast <ObjectDrawer *> (&*odf->makeIt ());
    odconn->setObjectDrawer (od, s_me);
  } else {
    // If they make a PhysicsObject before the object drawer factory
    // is set, or we are setting the odf to 0 just before shutting
    // down a window system, then we fall into this case. 
    odconn->setObjectDrawer (0, s_me);
  }
}

void
TheObjectDrawerFactory::registerConnection (SP<ObjectDrawerConnection> odc) {
  TheObjectDrawerFactoryData::makeIt ();
  Dynavec <ObjectDrawerConnection *> &conns = s_me->m_data->m_connections;
  int pos = conns.size();
  conns.push (odc);
  odc->setObjectDrawerPosition (pos);
  giveObjectDrawer (odc);
}

void
TheObjectDrawerFactory::deRegisterConnection (SP<ObjectDrawerConnection> odc) {
  Dynavec <ObjectDrawerConnection *> &conns = m_data->m_connections;
  int position = odc->getObjectDrawerPosition ();
  assert (conns [position] == odc);
  if (position < conns.size() - 1) {
    conns [position] = conns [conns.size() - 1];
    conns [position]->setObjectDrawerPosition (position);
  }
  conns [conns.size() - 1] = 0;
  conns.pop ();
}

void
TheObjectDrawerFactory::setTheObjectDrawerFactory
(SP<ConstantRecursiveSlotValue> odf)
{
  TheObjectDrawerFactoryData::makeIt ();
  s_me->m_data->m_odf = odf;
  Dynavec <ObjectDrawerConnection *> &conns = s_me->m_data->m_connections;
  for (int i = 0; i < conns.size(); i++) {
    giveObjectDrawer (conns [i]);
  }
}

void TheObjectDrawerFactory::registerObjectDrawer (const FString &className) {
  TheObjectDrawerFactoryData::makeIt ();
  s_me->m_data->m_objectDrawers.push (className);
}

Dynavec <FString> TheObjectDrawerFactory::allRegisteredObjectDrawers () {
  TheObjectDrawerFactoryData::makeIt ();
  return s_me->m_data->m_objectDrawers;
}
