// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

// Write a .d file for input into Brenner's MD Fortran code.
// Format is described in Structures/READMD in the tar file for Brenner's MD
// package.  What this file doesn't say is:
//
// All text past the header is read with format "*", so I don't have to get any
// columns in any particular places.

#ifndef SceneWriterConfiguration_h_included
#include "SceneWriterConfiguration.h"
#endif

#ifndef fstream_h_included
#include <fstream>
#define fstream_h_included
#endif

// For strerror.
#ifndef  string_h_included
#include <string.h>
#define string_h_included
#endif

// For errno.
#ifndef errno_h_included
#include <errno.h>
#define errno_h_included
#endif

#ifndef Action_h_included
#include "Action.h"
#endif

#ifndef TypedFactory_h_included
#include "TypedFactory.h"
#endif

#ifndef MemoryUtil_h_included
#include "MemoryUtil.h"
#endif

#ifndef SelectionManager_h_included
#include "SelectionManager.h"
#endif

#ifndef FloatSlotValue_h_included
#include "FloatSlotValue.h"
#endif

#ifndef Vec3SlotValue_h_included
#include "Vec3SlotValue.h"
#endif

#ifndef Vec3_h_included
#include "Vec3.h"
#endif

#ifndef PositiveFloatSlotValue_h_included
#include "PositiveFloatSlotValue.h"
#endif

#ifndef OL_h_included
#include "OL.h"
#endif

#ifndef TopLevel_h_included
#include "TopLevel.h"
#endif

#ifndef BoringAtom_h_included
#include "BoringAtom.h"
#endif

#ifndef AtomInfo_h_included
#include "AtomInfo.h"
#endif

#ifndef StubBrennerAtom_h_included
#include "StubBrennerAtom.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef FFloat_h_included
#include "FFloat.h"
#endif

namespace {
  FString timeName () {
    return "Start time (Femtoseconds)";
  }

  FString timeStepName () {
    return "Time step (Femtoseconds)";
  }

  FString cubeName () {
    return "Periodicity of space (Angstroms)";
  }

  class DSceneWriterConfiguration
    : public SceneWriterConfiguration
  {
  public:
    DSceneWriterConfiguration () {
      initializeSlot (timeName(), NEW (FloatSlotValue (0)));
      initializeSlot (timeStepName (), NEW (PositiveFloatSlotValue (0.5)));
      initializeSlot (cubeName (),
		      NEW (Vec3SlotValue (Vec3 (1000, 1000, 1000))));
    }
    FFloat getTime () const {
      return FloatSlotValue::getSlot (this, timeName ());
    }
    void setTime (FFloat val) {
      FloatSlotValue::setSlot (this, timeName (), val);
    }
    FFloat getTimeStep () const {
      return FloatSlotValue::getSlot (this, timeStepName ());
    }
    void setTimeStep (FFloat val) {
      FloatSlotValue::setSlot (this, timeStepName (), val);
    }
    Vec3 getCube () {
      return Vec3SlotValue::getSlot (this, cubeName());
    }
    void setCube (const Vec3 &val) {
      Vec3SlotValue::setSlot (this, cubeName(), val);
    }
  };
  typedef TypedFactory <DSceneWriterConfiguration, Action> TF;
  class DSceneWriter
    : public TF
  {
  public:
    DSceneWriter ()
      : TF ("DSceneWriter")
    {}
    SP<DSceneWriterConfiguration> typedDefaultConfiguration () const {
      return NEW (DSceneWriterConfiguration ());
    }
    SP<Action> makeIt (DSceneWriterConfiguration *conf) const {
      SP<Action> result = NEW (Action ());
      SP<SceneGraph> sg = conf->getTopLevel()->getSceneGraph();
      // The object indices to write out.
      Dynavec <int> serialOrder;
      if (conf->getSelectionOnly ()) {
	SelectionManager::getSelection (serialOrder);
      } else {
	for (int i = 0; i < sg->maxIndex (); i++) {
	  serialOrder.push (i);
	}
      }
      { // Filter out everything but the atoms.
        // Warn if any atoms are not H or C.
	int to = 0;
	const FString HSym = "H";
	const FString CSym = "C";
	for (int from = 0; from < serialOrder.size(); from++) {
	  const int s = serialOrder [from];
	  SP<const PhysicsObject> const o = sg->object(s);
	  if (o->isInstanceOf (BoringAtom::staticClassId)) {
	    SP<const BoringAtom> const ba =
	      static_cast <const BoringAtom *> (&*o);
	    const int num = ba->getSymbolNumber();
	    if (num != 1 & num != 6) {
	      result->setMessage (FString ("Atom ")+s+" is "+
				  AtomInfo::toSymbol (ba->getSymbolNumber())+
				  ", not H or C; writing it anyway");
	    }
	    serialOrder[to] = s;
	    to++;
	  } else {
	    // Skip the non-atom.
	  }
	}
	serialOrder.extendTo (to);
      }
      std::ofstream out (&*(conf->getFileName()));
      if (!out.is_open()) {
	result->setProblem (FString ("Failed to open output file ")+
			    conf->getFileName()+": "+strerror (errno));
      }
      // The header.
      out << " " << conf->getFileName () << std::endl;
      // Number of atoms.
      out << serialOrder.size() << std::endl;
      // Time and timestep in femtoseconds
      const FFloat timeStep = conf->getTimeStep ();
      FString line = "";
      FString error = "";
      line.fromFloat (conf->getTime (), 1, 20, 6, &error);
      line.fromFloat (timeStep, 26, 20, 6, &error);
      out << line << std::endl;
      if (error.size ()) {
        result->setProblem ("Could not format timestep line: "+error);
      }
      // The periodic cube.
      Vec3 cube = conf->getCube ();
      out << cube [0] << " " << cube [1] << " " << cube [2] << std::endl;
      // The Fortran code that reads these lines uses "*" as the format, so we
      // could almost write it out with variable width and be happy.  However,
      // our code that reads the files does fixed-width reads that line up with
      // the fixed-width writes that the Fortran does, so if we want to be able
      // to read our own output without falling down we need to do fixed-width
      // writes here.
      for (int i = 0; i < serialOrder.size() && !result->isProblem (); i++) {
	const int objectIndex = serialOrder [i];
	OL<BoringAtom> ba (sg, objectIndex);
	line = "";
	line.fromInt (i+1, 0, 5, &error);
	// Be sure to have a space between the columns, so Fortran's * format
	// will read correctly.
	line.fromInt (ba->getSymbolNumber(), 6, 4, &error);
	const FFloat *state = ba.objectState ();
	Vec3 where = BoringAtom::center (state);
	line.fromFloat (where [0], 11, 20, 6, &error);
	line.fromFloat (where [1], 31, 20, 6, &error);
	line.fromFloat (where [2], 51, 20, 6, &error);
	StubBrennerAtom::MotionFlag motion;
	if (ba->isInstanceOf (StubBrennerAtom::staticClassId)) {
	  SP<StubBrennerAtom> bra = (StubBrennerAtom *) (&*ba);
	  motion = bra->getMotionFlag ();
	} else {
	  motion = StubBrennerAtomConfiguration::DYNAMIC;
	}
	line.fromInt (motion, 72, 1, &error);
	if (error.size ()) {
	  result->setProblem ("Could not format line: "+error);
	} 
	out << line << std::endl;
	if (out.fail ()) {
	  result->setProblem ("I/O error on " + conf->getFileName ()+": "
			      + strerror (errno));
	}
      }
      // Write velocities.
      for (int i = 0; i < serialOrder.size() && !result->isProblem(); i++) {
	const int objectIndex = serialOrder [i];
	OL<BoringAtom> ba (sg, objectIndex);
	line = "";
	line.fromInt (i+1, 0, 5, &error);
	const FFloat *state = ba.objectState ();
	Vec3 vel = BoringAtom::velocity (state) * timeStep;
	line.fromFloat (vel[0], 5, 20, 6, &error);
	line.fromFloat (vel[1], 25, 20, 6, &error);
	line.fromFloat (vel[2], 45, 20, 6, &error);
	if (error.size()) {
	  result->setProblem ("Could not format line: "+error);
	}
	out << line << std::endl;
	if (out.fail ()) {
	  result->setProblem ("I/O error on " + conf->getFileName ()+": "
			      + strerror (errno));
	}
      }
      // Write three lumps of zeros.
      for (int j = 0; j < 3; j++) {
	for (int i = 0; i < serialOrder.size() && !result->isProblem(); i++) {
	  line = "";
	  line.fromInt (i+1, 0, 5, &error);
	  line.fromFloat (0, 6, 20, 6, &error);
	  line.fromFloat (0, 26, 20, 6, &error);
	  line.fromFloat (0, 46, 20, 6, &error);
	  if (error.size()) {
	    result->setProblem ("Could not format line: "+error);
	  }
	  out << line << std::endl;
	  if (out.fail ()) {
	    result->setProblem ("I/O error on " + conf->getFileName ()+": "
				+ strerror (errno));
	  }
	}
      }
      out.close ();
      if (!result->isProblem()) {
	if (out.fail ()) {
	  result->setProblem ("I/O error while closing "+
			      conf->getFileName()+": "+strerror (errno));
	}
      }
      return result;
    }
  };
  static const bool useless =
  (FactoryTable::store ("SceneWriter", NEW (DSceneWriter ())),
   true);
}
