// Fungimol - an extensible system for designing atomic-scale objects.
// Copyright (C) 2000 Tim Freeman
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
// 
// You should have received a copy of the GNU Library General Public
// License along with this library in the file COPYING.txt; if not,
// write to the Free Software Foundation, Inc., 59 Temple Place -
// Suite 330, Boston, MA 02111-1307, USA
//
// The author can be reached by email at tim@fungible.com, or by
// paper mail at:
//
// Tim Freeman
// 381 North Fernwood Circle
// Sunnyvale, CA 94085-3059
//

#ifndef FString_h_included
#define FString_h_included

#ifndef Refcount_h_included
#include "Refcount.h"
#endif

#ifndef Dynavec_h_included
#include "Dynavec.h"
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef FFloat_h_included
#include "FFloat.h"
#endif

#ifndef Hashable_h_included
#include "Hashable.h"
#endif

// Defines istream, std::ostream.  g++ 3.0.1 won't accept a forward declaration of
// std::ostream and istream here.  Tim Freeman  1 Sep 2001.
#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

class FString : public Hashable {
  Dynavec <char> m_chars;
  inline void FString::check () const {
    assert (m_chars.size() >= 1);
    assert ('\0' == m_chars[m_chars.size()-1]);
  }
public:
  FString ();
  virtual ~FString ();
  FString (char c);
  FString (const char *c);
  FString (const FString &s);
  FString (int i);
  FString (double i);
  // Chronic problem -- any pointer type converts to bool which converts to
  // FString.  The "explicit" keyword on page 423 of the C++ ARM is supposed to
  // solve exactly this problem, and it does!  Yay!
  explicit FString (bool i);
  FString & operator= (const FString &s);
  // On error, return 0.0 and set *error (if specified and nonzero) to a
  // nonempty FString describing what went wrong.
  // On success, return the floating point number and leave *error
  // unmodified.  This way, error can be used to accumulate errors
  // from several calls.  If error is not specified, we'll abort with a message
  // if parsing fails.
  // A blank FString is an error.  FIXME This is misimplemented now.
  // Fixing would require changes to the PDB reader and is not a high priority
  // right now.
  FFloat toFloat (int start = 0, FString *error = 0) const;
  // chars is the maximum field width.
  FFloat toFloat (int start, int chars, FString *error = 0) const;
  // Same, but for integers.
  int toInt (int start = 0, FString *error = 0) const;
  int toInt (int start, int chars, FString *error = 0) const;
  // Whether the specified substring is entirely whitespace.
  //
  // FIXME We really need a constant-time substring operation.
  //
  // All these FString operations that really work on substrings are silly.
  // Array references into strings would have to return an array-reference
  // object, which would copy the destination FString if it has constant-time
  // substrings pointing into it.  Or maybe we just don't allow mutable
  // references into strings and fix the code that wants it to do something
  // else.  Hmm, converting FString to char* would have to be disallowed, since
  // we couldn't guarantee the null termination of the char* if it was a
  // substring reference to the middle of some longer FString.
  //
  // FIXME isBlank isn't implemented at all.
  bool isBlank (int start, int chars) const;
  bool isBlank (int start = 0) const;
  // Convert a FFloat to ascii with the given width and insert it into the
  // FString.  I would like to use printf for this, but the manual entry says
  // "In no case does a non-existent or small field width cause truncation of a
  // field; if the result of a conversion is wider than the field width, the
  // field is expanded to contain the conversion result."  PDB files really
  // have fixed-width fields, so direct use of printf isn't good enough.
  // Places is the preferred number of digits after the decimal point, although
  // if there isn't room you'll get less than that.  If the integer part can't
  // fit, you should get an error.  If the FString has to be extended so it has
  // characters at the required positions, we extend with spaces.
  void fromFloat (FFloat val, int start, int chars, int places,
		  FString *error = 0);
  // Likewise for integers.
  void fromInt (int val, int start, int chars, FString *error = 0);
  // Copy the given FString into a substring of this.  Left justify and pad with
  // spaces.  Get an error if it doesn't fit.
  void fromString (const FString &val, int start, int chars, FString *error = 0);
  // Like fromString, except right justify.
  void fromStringRight (const FString &val, int start, int chars,
			FString *error = 0);
  // Returns a substring of "this" starting at from, and extending to the end
  // of the FString or for chars characters, whichever is shorter.
  FString substring (int from, int chars) const;
  // x.matches (y, m) is equivalent to
  // x.substring (y, strlen (m)) == m
  // but faster because it avoids the strlen and more importantly it avoids
  // allocating a new FString for the substring.
  bool matches (int position, const char *match);
  Hashable::HASH hash () const;
  // Next one is for the Hashable interface.
  FString *copy () const;
  friend FString operator+ (const FString &s1, const FString &s2);
  const FString &operator+= (const FString &s);
  const FString &operator+= (char ch);
  bool operator== (const Hashable &s) const; 
  friend bool operator== (const FString &s1, const FString &s2);
  friend bool operator== (const FString &s, const char *cp);
  friend bool operator== (const char *cp, const FString &s);
  bool operator!= (const FString &s) const;
  friend bool operator< (const FString &s1, const FString &s2);
  char &operator[] (int i);
  // *s is equivalent to s[0], except you're allowed to do *s even if
  // the FString has length 0.  Usual idiom is &*s to convert to a
  // char * pointer, which will be null terminated.
  char &operator * ();
  const char &operator * () const;
  // Have const char & immediately below instead of just char because I want to
  // be able to take &foo[n] and get a char*.
  const char &operator[] (int i) const;
  // Number of characters in the FString.  There's a trailing null
  // hidden at the end, but that's just a representation detail, so
  // it's none of the user's business and it isn't counted in the
  // size.  Putting it inline because it tends to be called repeatedly if the
  // nonemptyness of a FString is used as an error indicator.
  int size () const {
    check ();
    return m_chars.size() - 1;
  }

  FString upcase () const;
  friend std::ostream &operator<< (std::ostream &o, const FString& s);
  // Next one reads the newline and discards it.
  friend std::istream &operator>> (std::istream &i, FString &s);
};

#endif
