// This is the glue that makes the Brenner force field into a Fungimol atom
// plugin.

#ifndef RLIST_h_included
#include "rlist.h"
#endif

// brenc.h declares ljparam_init_kt, among others.
#ifndef BRENC_h_included
#include "brenc.h"
#endif

#ifndef FUNGIMOLAPI_h_included
#include "fungimolapi.h"
#endif

#ifndef BrennerAtom_h_included
#include "BrennerAtom.h"
#endif

#ifndef BRENNER_h_included
#include "brenner.h"
#endif

#ifndef State_h_included
#include "State.h"
#endif

#ifndef AtomConfiguration_h_included
#include "AtomConfiguration.h"
#endif

#ifndef AtomInfo_h_included
#include "AtomInfo.h"
#endif

#ifndef KT_h_included
#include "kt.h"
#endif

#ifndef myassert_h_included
#include "myassert.h"
#endif

#ifndef XMASS_h_included
#include "xmass.h"
#endif

#ifndef SceneGraph_h_included
#include "SceneGraph.h"
#endif

#ifndef SceneGraphPOInterface_h_included
#include "SceneGraphPOInterface.h"
#endif

#ifndef XMMS_h_included
#include "xmms.h"
#endif

#ifndef RSLJ_h_included
#include "rslj.h"
#endif

#ifndef SP_h_included
#include "SP.h"
#endif

#ifndef AtomCache_h_included
#include "AtomCache.h"
#endif

#ifndef TypedFactory_h_included
#include "TypedFactory.h"
#endif

#ifndef FactoryTable_h_included
#include "FactoryTable.h"
#endif

#ifndef FString_h_included
#include "FString.h"
#endif

#ifndef iostream_h_included
#include <iostream>
#define iostream_h_included
#endif

namespace {
  // We'll pass this to the C functions that do the real physics.
  State *s_state = 0;
  
  void deallocate () {
    if (s_state != 0) {
      delete s_state;
    }
    deallocate_caguts ();
    deallocate_rangetree ();
  }

  // We have to re-allocate some memory if the scene graph garbage collects,
  // so we have to remember the number of garbage collections that happened
  // last time so we can see if it's different this time.
  int gccount = -1;
#ifndef NDEBUG
  // Keep track of the first object index that's a BrennerAtom, so we can
  // assert that the scene graph is really sorted.
  int initialIndex = -1;
#endif
  int lastIndex = -1;
} // namespace  

static void misc_init () {
  // Saying this because of the license on bren2.c.  I probably don't really
  // have to, but it's in keeping with the other messages that Fungimol prints
  // when it starts.
  std::cout << "This product includes software developed by Zyvex LLC." << std::endl;
  static_initkt ();
  static_initxmass ();
  s_state = NEW (State ());
  MemoryUtil::registerDeallocator (deallocate);
  // FIXME It would be better to take rll as an input in the configuration,
  // but that would require taking a subclass of AtomConfiguration here.
  // rll is the slop to insert into neighboring so we don't have to recompute
  // neighbors for every timestep.
  s_state->rll = 0.5;
  // init_c calls mtable.
  init_c ();
  // Wire in the lennard-jones parameters for carbon here.
  // Beware that the order of the arguments is the opposite of the order of
  // the values in input.d.
  ljparam_init_kt (kt[6], 2.28, 51.2);
  // And hydrogen.  Changing this in input.d causes us to get different
  // results from the Fortran version, so I don't want to play with this
  // right now. 
  // ljparam_init_kt (kt[1], 2.81, 8.6);
  // PARAM calls ljparam.
  PARAM (ktmax, s_state->rll);
}

static bool useless = (misc_init(), true);

static inline int safeSymbolNumberToKtype (int symbolNumber) {
  if (symbolNumber >= AtomInfo::maxNumber ||
      symbolNumber >= MAX_ATOMNO ||
      symbolNumber < 1) {
    return 1;
  } else {
    const int result = kt [symbolNumber];
    assert (result > 0);
    return result;
  }
}

// We need static_initkt() to be in a static initializer because
// safeSymbolNumberToKtype below uses it.  Doing it the first time only inside
// the constructor is too late.
BrennerAtom::BrennerAtom (AtomConfiguration *b, const Factory *f)
  : BoringAtom (b, f),
    thisKtype (safeSymbolNumberToKtype (b->getSymbolNumber ()))
{}

const char *const BrennerAtom::staticClassId = "BrennerAtom";

const char *BrennerAtom::classId () const {
  return BrennerAtom::staticClassId;
};

bool BrennerAtom::isInstanceOf (const char *classId) const {
  if (staticClassId == classId) {
    return true;
  } else {
    return BoringAtom::isInstanceOf (classId);
  }
}

bool BrennerAtom::exempt () const {
  return false;
}

bool BrennerAtom::isRelevantNeighbor (const PhysicsObject *o, int myIndex,
                                      int neighborIndex) const {
  (void) o; (void) myIndex; (void) neighborIndex;
  return false;
}

// Taking this out for now.  See comments in BrennerAtom.h.
//bool BrennerAtom::lessThan (const PhysicsObject *other) const {
//  const char *otherClassId = other->classId ();
//  if (otherClassId == BrennerAtom::staticClassId) {
//    return getSymbolNumber () <
//      static_cast <const BrennerAtom *> (other)->getSymbolNumber ();
//  } else {
//    return compareKeyLessThan (BrennerAtom::staticClassId, otherClassId);
//  }
//}

int BrennerAtom::doPhysics (const SceneGraphPOInterface *const sg,
                            const Float *const state,
                            Float *const deriv,
                            const int myObjectIndex) const {
  bool reNeighbor = false;
  s_state->state = state;
  s_state->deriv = deriv;
  if (SceneGraph::garbageCollectionCount () != gccount) {
    gccount = SceneGraph::garbageCollectionCount ();
#ifndef NDEBUG
    initialIndex = myObjectIndex;
#endif
    lastIndex = myObjectIndex;
    for (;;) {
      if (lastIndex >= sg->maxIndex()) break;
      assert (sg->object (lastIndex));
      if (!sg->object(lastIndex)->isInstanceOf (BrennerAtom::staticClassId)) {
        break;
      }
      lastIndex++;
    }
    s_state->setRange (myObjectIndex, lastIndex);
    // fill in State.noa
    {
      Dynavec <int> &noa = s_state->noa;
      s_state->noa.extendTo (0);
      for (int i = myObjectIndex; i < lastIndex; i++) {
        int newKtype =
          (static_cast <const BrennerAtom *> (sg->object(i)))->ktype();
        if (newKtype >= noa.size()) {
          noa.extendTo (newKtype + 1);
        }
        noa[newKtype]++;
      }
    }
    s_state->sg = sg;
    reNeighbor = true;
    // setAllPrevToCurrent requires that s_state->state has been set.
    setAllPrevToCurrent (s_state);
  }
  // Check that the scene graph is really sorted.
  assert (myObjectIndex == initialIndex);
  // Check that there's only one scene graph.
  assert (s_state->sg == sg);
  // Initialize the forces to zero
  const int pastLastIndex = s_state->getPastLastIndex ();
  for (int i = myObjectIndex; i < pastLastIndex; i++) {
    vector &v = s_state->force (i);
    v.x = 0;
    v.y = 0;
    v.z = 0;
  }
  if (reNeighbor || (1 == choose_lj (s_state))) {
    // PARAM calls ljparam which initializes rslj and xmms.
    make_neighbor_lists_tree (s_state, ljNeighbors (s_state), rslj, xmms);
    /* zeroxmms is initialized by the compiler to all zeros.  For the
       purposes of neighboring we want the minimum cutoff distance to be
       zero. */
    static Float zeroxmms [NTYPES+1][NTYPES+1]; 
    /* RLIST was initialized when we called PARAM. */
    make_neighbor_lists_tree (s_state, caNeighbors (s_state),
                              RLIST, zeroxmms);
  }
  caguts (s_state);
  ljPure (s_state);
  for (int i = myObjectIndex; i < pastLastIndex; i++) {
    const int stateStart = i * 6;
    // Velocity is the derivative of position.
    deriv [stateStart] += state [stateStart+3];
    deriv [stateStart+1] += state [stateStart+4];
    deriv [stateStart+2] += state [stateStart+5];
    // Divide force by mass to get acceleration.
    // We could save a few cycles here by taking advantage of the fact that the
    // scene graph is sorted by atomic number.  We don't need to fetch the
    // atomic number every time or index into xmass every time.  I don't have
    // test cases for the silicon and germanium code, so it's probably broken,
    // so we'd only have to pay attention to hydrogen and carbon.
    const Float mass = emass [static_cast <const BrennerAtom *>
                             (s_state->sg->object(i))->ktype ()];
    deriv [stateStart+3] += floatCheckNAN (s_state->force(i).x / mass);
    deriv [stateStart+4] += floatCheckNAN (s_state->force(i).y / mass);
    deriv [stateStart+5] += floatCheckNAN (s_state->force(i).z / mass);
  }
  return pastLastIndex;
}

void my_exit(int code)
{
#ifndef NDEBUG
  abort ();
#endif
  exit(code);
}

class BrennerAtom::BrennerAtomFactory
  : public TypedFactory <AtomConfiguration, BrennerAtom>
{
  static SP<AtomCache <BrennerAtom> > s_cache;
  static void deallocate () {
    s_cache = 0;
  }
  static const bool s_useless;
public:
  BrennerAtomFactory ()
    : TypedFactory <AtomConfiguration, BrennerAtom> ("BrennerAtom")
  {}
  SP<AtomConfiguration> typedDefaultConfiguration () const {
    return NEW (AtomConfiguration ());
  }
  SP<BrennerAtom> makeIt (AtomConfiguration *ai) const {
    SP<BrennerAtom> &ba = s_cache->cacheref (ai);
    if (!ba) {
      ba = NEW (BrennerAtom (ai, this));
    }
    return ba;
  }
};

SP<AtomCache<BrennerAtom> > BrennerAtom::BrennerAtomFactory::s_cache =
NEW (AtomCache <BrennerAtom> ());

namespace {
  FString atomProtocol () {
    return "Atom";
  }
}

const bool BrennerAtom::BrennerAtomFactory::s_useless =
  (FactoryTable::store (atomProtocol (), NEW (BrennerAtomFactory ())),
   MemoryUtil::registerDeallocator (BrennerAtom::BrennerAtomFactory::deallocate),
   true);

