/* bvector.h - Copyright (c) 1998 Zyvex LLC.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef bvector_h_included
#define bvector_h_included

/* This file used to be called vector.h, but Walter Vannini points out
   (correctly, IMO) that this is likely to conflict with preexisting files in
   various C compilers, so I'm renaming it to bvector.h.  Tim Freeman 14 Aug
   2001 */

/* The next one is required for pow. */
#ifndef MATH_H
#include <math.h>
#define MATH_H
#endif

/* Float and Double are supposed to be defined on the C command line. */
typedef struct { Float x,y,z; } vector;
typedef struct { Double x,y,z; } dvector;

/* Vectors used to require a subroutine call to use.  Now I inline them in this
   header file.  The duration of the timeruns changed from 4.90 seconds to
   4.88.  I had hoped for more dramatic results, but given that there were any
   results at all I'm unwilling to un-inline.  Inlining vectors made a much
   larger difference in Fungimol, I think because I used the analogous vector
   class more there.  Tim Freeman 29 Jul 2000. */

/* Add two vectors */
static inline vector plus( vector i, vector j )
{
    vector sum;
    sum.x = i.x + j.x;
    sum.y = i.y + j.y;
    sum.z = i.z + j.z;
    return sum;
}

static inline dvector dplus( dvector i, vector j )
{
    dvector sum;
    sum.x = i.x + j.x;
    sum.y = i.y + j.y;
    sum.z = i.z + j.z;
    return sum;
}

static inline vector minus ( vector i, vector j )
{
    vector sum;
    sum.x = i.x - j.x;
    sum.y = i.y - j.y;
    sum.z = i.z - j.z;
    return sum;
}

/* Note that subtracting two double-precision vectors gives you a
   single-precision vector.  This is because we use double-precision vectors
   for position, and I don't think we need the precision for the difference
   between two vectors. */
static inline vector ddminus ( dvector i, dvector j )
{
    vector sum;
    sum.x = i.x - j.x;
    sum.y = i.y - j.y;
    sum.z = i.z - j.z;
    return sum;
}

/* If we switch to C++, we don't have to keep straight the difference in
   meaning between plus and sum, and we wouldn't have to have the different
   spellings of minus and plus above.  But IMO it isn't worth switching to C++
   only for that.  Tim Freeman 30 Jul 2000. */
static inline vector sum( Float i, vector j )
{
    vector result;
    result.x = i + j.x;
    result.y = i + j.y;
    result.z = i + j.z;
    return result;
}

/* Find the product of a scalar and a vector. */
static inline vector product( Float i, vector j )
{
    vector vec;
    vec.x = i*j.x; 
    vec.y = i*j.y; 
    vec.z = i*j.z;
    return vec;
}

/* Raise a vector to a power elementwise. */

static inline vector power( Float i, vector j )
{
    vector vec;
    vec.x = pow(j.x, i);
    vec.y = pow(j.y, i);
    vec.z = pow(j.z, i);
    return vec;
}

/* Find the dot product of two vectors. */
static inline Float dot( vector i, vector j )
{
    return ( i.x*j.x + i.y*j.y + i.z*j.z );
}

/* Find the cross product of two vectors */
static inline vector cross( vector b, vector c)
{
    vector vec;
    vec.x = b.y*c.z - c.y*b.z;
    vec.y = b.x*c.z - c.x*b.z;
    vec.z = b.x*c.y - c.x*b.y;
    return vec;
}

#endif
