#!/usr/local/sbin/perl -w
use strict;

my $default_threshold = 0.002;
sub usage () {
    die << "EOM";
Usage: $0 file1.d file2.d [<threshold>]
Scan through these files, looking for floating point numbers
preceeded with "CMP".  When we find them, check that they're in the
same place in the two files, and then compare the two corresponding
numbers for nearness.  If they differ by more than <threshold>, stop
and report an error.

We only allow one CMP per line.

The default value for threshold is $default_threshold.
EOM
}

my $threshold;

sub docmp ($$) {
    my ($n1, $n2) = @_;
    if (abs ($n1 - $n2) > $threshold) {
	die "Significantly different numbers $n1 and $n2 found";
    }
}

sub main () {
    usage () unless 3 == @ARGV || 2 == @ARGV;
    my ($file1, $file2) = @ARGV;
    if (3 == @ARGV) {
	$threshold = $ARGV[2];
    } else {
	$threshold = $default_threshold;
    }
    open (FILE1, "<$file1") || die "Can't read $file1: $!";
    open (FILE2, "<$file2") || die "Can't read $file2: $!";
    for (;;) {
	my $line1 = <FILE1>;
	my $line2 = <FILE2>;
	last if !defined ($line1) && !defined ($line2);
	die "The files have different numbers of lines"
	    if !defined ($line1) || !defined ($line2);
	die "Too many CMP's on this line of $file1: $line1"
	    if $line1 =~ m!CMP.*CMP!;
	die "Too many CMP's on this line of $file2: $line2"
	    if $line2 =~ m!CMP.*CMP!;
	if ($line1 =~ m!CMP\s*([^\s]+)!) {
	    my $num1 = $1;
	    if ($line2 =~ m!CMP\s*([^\s]+)!) {
		my $num2 = $1;
		my $diff = sprintf ("%f",$num2 - $num1);
		$line1 =~ m!^(.*)CMP\s*([^\s]+)(.*)$! || die "Impossible";
		my ($before1, $during1, $after1) = ($1, $2, $3);
		$line2 =~ m!^(.*)CMP\s*([^\s]+)(.*)$! || die "Impossible";
		my ($before2, $during2, $after2) = ($1, $2, $3);
		if ($before1 ne $before2 || $after1 ne $after2) {
		    die ("The text bits are different, never mind the numbers:\n".
			 "   ${line1}and\n   $line2");
		}
                print "${before1}CMP $during1 ($diff)$after1\n";
		docmp ($num1, $num2);
	    } else {
		die "CMP on this line of $file1 with none in $file2: $line1";
	    }
	} else {
	    if ($line2 =~ m!CMP\s*([^\s]+)!) {
		die "CMP on this line of $file2 with none in $file1: $line2";
	    } else {
		print $line1;
	    }
	} 
    }
    close (FILE1) || die "Can't close $file1: $!";
    close (FILE2) || die "Can't close $file2: $!";
    print "Files are close enough.\n";
}

main ();
